@@ -23,10 +23,17 @@ import (
 	"path/filepath"
 	"testing"
 
+	"github.com/prometheus/prometheus/tsdb/fileutil"
+	"go.uber.org/goleak"
+
 	client_testutil "github.com/prometheus/client_golang/prometheus/testutil"
 	"github.com/prometheus/prometheus/util/testutil"
 )
 
+func TestMain(m *testing.M) {
+	goleak.VerifyTestMain(m)
+}
+
 // TestWALRepair_ReadingError ensures that a repair is run for an error
 // when reading a record.
 func TestWALRepair_ReadingError(t *testing.T) {
@@ -46,8 +53,8 @@ func TestWALRepair_ReadingError(t *testing.T) {
 			8,
 		},
 		// Ensures that the page buffer is big enough to fit
-		// an entire page size without panicing.
-		// https://github.com/prometheus/prometheus/tsdb/pull/414
+		// an entire page size without panicking.
+		// https://github.com/prometheus/tsdb/pull/414
 		"bad_header": {
 			1,
 			func(f *os.File) {
@@ -300,7 +307,7 @@ func TestCorruptAndCarryOn(t *testing.T) {
 		err = w.Repair(corruptionErr)
 		testutil.Ok(t, err)
 
-		// Ensure that we have a completely clean slate after reapiring.
+		// Ensure that we have a completely clean slate after repairing.
 		testutil.Equals(t, w.segment.Index(), 1) // We corrupted segment 0.
 		testutil.Equals(t, w.donePages, 0)
 
@@ -377,7 +384,7 @@ func TestSegmentMetric(t *testing.T) {
 }
 
 func TestCompression(t *testing.T) {
-	boostrap := func(compressed bool) string {
+	bootstrap := func(compressed bool) string {
 		const (
 			segmentSize = pageSize
 			recordSize  = (pageSize / 2) - recordHeaderSize
@@ -399,17 +406,19 @@ func TestCompression(t *testing.T) {
 		return dirPath
 	}
 
-	dirCompressed := boostrap(true)
+	dirCompressed := bootstrap(true)
 	defer func() {
 		testutil.Ok(t, os.RemoveAll(dirCompressed))
 	}()
-	dirUnCompressed := boostrap(false)
+	dirUnCompressed := bootstrap(false)
 	defer func() {
 		testutil.Ok(t, os.RemoveAll(dirUnCompressed))
 	}()
 
-	uncompressedSize := testutil.DirSize(t, dirUnCompressed)
-	compressedSize := testutil.DirSize(t, dirCompressed)
+	uncompressedSize, err := fileutil.DirSize(dirUnCompressed)
+	testutil.Ok(t, err)
+	compressedSize, err := fileutil.DirSize(dirCompressed)
+	testutil.Ok(t, err)
 
 	testutil.Assert(t, float64(uncompressedSize)*0.75 > float64(compressedSize), "Compressing zeroes should save at least 25%% space - uncompressedSize: %d, compressedSize: %d", uncompressedSize, compressedSize)
 }
